/*
 * imagebuf.c
 *
 *  Created on: 2013/04/29
 *  Copyright(C) 2013 Ryoichi Yokomizu
 *  All rights reserved
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "imagebuf.h"
#include "bmpfile.h"

/*#define IMAGEBUF_DEBUG*/

/* 外部変数の参照 */
extern const unsigned char g_FontData5x8[][8];

/* 定数定義 */
#define IMAGEBUF_MAGICID	('F' | ('B' << 8))
#define BIWIDTHBYTES(w)		(((w) * 3 + 3) & 0xfffffffc)
#define BYTES_PER_PIXEL(b)	(((b) + 7) / 8)

#define IMG_FONT_WIDTH		5
#define IMG_FONT_HEIGHT		8
#define IMG_FONT_MAXMAG		16

#if defined(__sh3__)
  #define UNCACHED(a)	(((unsigned long)(a) & 0x1fffffff) | 0xa0000000)
  #define IMG_USE_DMAC
#else
  #define	UNCACHED(a) a
#endif

/* 内部変数の定義 */
typedef struct tagImageBufferSt {
	unsigned short magicid;
	unsigned short width;
	unsigned short height;
	unsigned short bitsperpixel;
	unsigned short bytesperpixel;
	unsigned short mode;
	unsigned long size;
} IMAGEBUF_ST, *PIMAGEBUF_ST;

/* 内部関数の定義 */
static void img_checkpos(short *x, short *y, short width, short height);
static void img_hline_rgb888(unsigned char *ptr, short count, unsigned char r, unsigned char g, unsigned char b);
static void img_hline_rgb565(unsigned short *wptr, short count, unsigned short wval);
static void img_plot4points(HIMAGEBUF hBuffer, short cx, short cy, short x, short y, unsigned char r, unsigned char g, unsigned char b);

HIMAGEBUF IMG_CreateOffscreenBuffer(void)
{
	PIMAGEBUF_ST pheader = 0;
	unsigned long bufsize;

	bufsize = sizeof(IMAGEBUF_ST) + (LCDC_GetScreenWidth() * BYTES_PER_PIXEL(LCDC_GetBitsPerPixel())) * LCDC_GetScreenHeight();
	pheader = (PIMAGEBUF_ST)malloc(bufsize);
	if (pheader != 0) {
		pheader->magicid = IMAGEBUF_MAGICID;
		pheader->width  = LCDC_GetScreenWidth();
		pheader->height = LCDC_GetScreenHeight();
		pheader->bitsperpixel = LCDC_GetBitsPerPixel();
		pheader->bytesperpixel = BYTES_PER_PIXEL(LCDC_GetBitsPerPixel());
		pheader->mode = LCDC_GetImageFormat();
		pheader->size = (LCDC_GetScreenWidth() * BYTES_PER_PIXEL(LCDC_GetBitsPerPixel())) * LCDC_GetScreenHeight();
#ifdef IMAGEBUF_DEBUG
		printf("IMAGEBUF: CreateBuffer width=%d, height=%d, bytes=%d, size=%d, ptr=%08x, start=%08x\n",
				pheader->width, pheader->height, pheader->bytesperpixel, pheader->size,
				pheader, (unsigned char *)pheader + sizeof(pheader));
#endif
	}

	return (void *)pheader;
}

HIMAGEBUF IMG_CreateBuffer(short width, short height)
{
	PIMAGEBUF_ST pheader = 0;
	unsigned long bufsize;

	if ((width < 0) || (width > MAX_BUFFER_WIDTH)) {
		return 0;
	}

	if ((height < 0) || (height > MAX_BUFFER_HEIGHT)) {
		return 0;
	}

	bufsize = sizeof(IMAGEBUF_ST) + (width * BYTES_PER_PIXEL(LCDC_GetBitsPerPixel())) * height;
	pheader = (PIMAGEBUF_ST)malloc(bufsize);
	if (pheader != 0) {
		pheader->magicid = IMAGEBUF_MAGICID;
		pheader->width  = width;
		pheader->height = height;
		pheader->bitsperpixel = LCDC_GetBitsPerPixel();
		pheader->bytesperpixel = BYTES_PER_PIXEL(LCDC_GetBitsPerPixel());
		pheader->mode = LCDC_GetImageFormat();
		pheader->size = (width * BYTES_PER_PIXEL(LCDC_GetBitsPerPixel())) * height;
#ifdef IMAGEBUF_DEBUG
		printf("IMAGEBUF: CreateBuffer width=%d, height=%d, bytes=%d, size=%d, ptr=%08x, start=%08x\n",
				pheader->width, pheader->height, pheader->bytesperpixel, pheader->size,
				pheader, (unsigned char *)pheader + sizeof(pheader));
#endif
	}

	return (void *)pheader;
}

void IMG_DeleteBuffer(HIMAGEBUF hBuffer)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {
			free((void *)pheader);
		}
	}
}

void IMG_TransferToScreen(HIMAGEBUF hBuffer)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	unsigned char *ptr;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {
			ptr = (unsigned char *)UNCACHED((unsigned char *)pheader + sizeof(IMAGEBUF_ST));
#ifdef IMAGEBUF_DEBUG
			printf("IMAGEBUF: TransterToScreen: Start:%08x, Size:%0d\n", ptr, pheader->size);
#endif
			LCDC_BurstWrite(0, ptr, pheader->size);	/* Offset:0から転送開始 */
		}
	}
}

void IMG_BitBltToScreen(HIMAGEBUF hBuffer, short x1, short y1, short width, short height, short srcx, short srcy)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	unsigned char *ptrsrc;
	unsigned short y, count, offset_src, offset_dst;
	unsigned long adrdst;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {
			/* 座標のチェック */
			if (x1 < 0) {
				width -= (-x1);
				if (width < 0) {
					return;
				}
				srcx  += (-x1);
				x1 = 0;
			}
			else if ((x1 + width) > LCDC_GetScreenWidth()) {
				width = LCDC_GetScreenWidth() - x1;
				if (width <= 0) {
					return;
				}
			}
			else if (x1 >= LCDC_GetScreenWidth()) {
				return;
			}
			else {
				;
			}
			if (y1 < 0) {
				height -= (-y1);
				if (height < 0) {
					return;
				}
				srcy += (-y1);
				y1 = 0;
			} else if ((y1 + height) > LCDC_GetScreenHeight()) {
				height = LCDC_GetScreenHeight() - y1;
				if (height <= 0) {
					return;
				}
			} else if (y1 >= LCDC_GetScreenHeight()) {
				return;
			}
			else {
				;
			}

			adrdst = (LCDC_GetScreenWidth() * y1 + x1) * BYTES_PER_PIXEL(LCDC_GetBitsPerPixel());

			ptrsrc = (unsigned char *)pheader + sizeof(IMAGEBUF_ST)
					+ (pheader->width * srcy + srcx) * pheader->bytesperpixel;
			ptrsrc = (unsigned char *)UNCACHED(ptrsrc);

			offset_src = pheader->width * pheader->bytesperpixel;
			offset_dst = LCDC_GetScreenWidth() * BYTES_PER_PIXEL(LCDC_GetBitsPerPixel());

			count = width * pheader->bytesperpixel;

			for (y = 0; y < height; y++) {
				LCDC_BurstWrite(adrdst, ptrsrc, count);
				ptrsrc += offset_src;
				adrdst += offset_dst;
			}
		}
	}
}

/* 描画関連 */
int IMG_LoadBMP(HIMAGEBUF hBuffer, const char *filename)
{
	FILE* fp = NULL;
	int err = 0;
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	BITMAPFILEHEADER bf;
	BITMAPINFOHEADER bih;

	if (pheader != 0) {
		if (pheader->magicid != IMAGEBUF_MAGICID) {
			err = 1;
		}
	}

	if (err == 0) {
		fp = fopen((char *)filename, "rb");
		if (fp == NULL) {
#ifdef IMAGEBUF_DEBUG
			printf("Cannot open file: %s\n", filename);
#endif
			err = 2;
		}
	}

	if (err == 0) {
		if (fread((char *)&bf, 1, sizeof(bf), fp) != sizeof(bf)) {
#ifdef IMAGEBUF_DEBUG
			printf("Cannot read file: %s\n", filename);
#endif
			err = 3;
		}
	}

	if (err == 0) {
		if (bf.bfType != 0x4d42) {
#ifdef IMAGEBUF_DEBUG
			printf("Illegal file format: 4d42h => %04xh\n", bf.bfType);
#endif
			err = 4;
		}
	}

	if (err == 0) {
		if (fread((char *)&bih, 1, sizeof(bih), fp) != sizeof(bih)) {
#ifdef IMAGEBUF_DEBUG
			printf("Cannot read file: %s\n", filename);
#endif
			err = 5;
		}
	}

	if (err == 0) {
		if ((bih.biBitCount != 24) || (bih.biWidth != pheader->width) || (bih.biHeight != pheader->height)) {
#ifdef IMAGEBUF_DEBUG
			printf("Illegal bitmap format (biBitCount=%d, biWidth=%d, biHeight=%d.\n", bih.biBitCount, bih.biWidth, bih.biHeight);
#endif
			err = 6;
		}
	}

	if (err == 0) {
		long readsize;
		int x, y;
		unsigned char *pbuf = 0, *psrc = 0, *ptemp;
		unsigned char r, g, b;
		unsigned short *prgb565;
////////unsigned short bitsperpixel = LCDC_GetBitsPerPixel();
		unsigned short bitsperpixel = pheader->bitsperpixel;

		readsize = BIWIDTHBYTES(pheader->width);

		if (bitsperpixel == 24) {
			for (y = 0; y < pheader->height; y++) {
				pbuf = (unsigned char *)pheader + sizeof(IMAGEBUF_ST)
					 + pheader->width * pheader->bytesperpixel * (pheader->height - 1 - y);
				pbuf = (unsigned char *)UNCACHED(pbuf);
				if (fread((char *)pbuf, 1, readsize, fp) != readsize) {
#ifdef IMAGEBUF_DEBUG
					printf("Cannot read bitmap: %s %dline\n", filename, y);
#endif
					err = 7;
					break;
				}
			}
		}
		else if (bitsperpixel == 16) {
			psrc = (unsigned char *)malloc(readsize);
			if (psrc != 0) {
				for (y = 0; y < pheader->height; y++) {
					if (fread((char *)psrc, 1, readsize, fp) != readsize) {
#ifdef IMAGEBUF_DEBUG
						printf("Cannot read bitmap: %s %dline\n", filename, y);
#endif
						err = 7;
						break;
					}
					prgb565 = (unsigned short *)((unsigned char *)pheader + sizeof(IMAGEBUF_ST)
							+ pheader->width * pheader->bytesperpixel * (pheader->height - 1 - y));
					prgb565 = (unsigned short *)UNCACHED(prgb565);
					ptemp = psrc;
					for (x = 0; x < pheader->width; x++) {
						b = *ptemp++ >> 3;	/* b: 5bit */
						g = *ptemp++ >> 2;	/* g: 6bit */
						r = *ptemp++ >> 3;	/* r: 5bit */
						*prgb565++ = (r << 11) | (g << 5) | b;
					}
				}
				free((void *)psrc);
			}
		}
		else {
#ifdef IMAGEBUF_DEBUG
			printf("Cannot support such color mode.\n");
			err = 7;
#endif
		}
	}

	if (fp != NULL) {
		close(fp);
	}

	return err;
}

void IMG_Fill(HIMAGEBUF hBuffer, unsigned char r, unsigned char g, unsigned char b)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	unsigned long count;
	unsigned char *ptrsrc;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {

			ptrsrc = (unsigned char *)pheader + sizeof(IMAGEBUF_ST);	/* 転送元アドレス */

			/* 転送元のデータを作る */
			switch(pheader->mode) {
			case LCDC_MODE_RGB888:
#ifdef IMG_USE_DMAC
				ptrsrc = (unsigned char *)UNCACHED(ptrsrc);
				*ptrsrc = b;
				*(ptrsrc + 1) = g;
				*(ptrsrc + 2) = r;
				*(ptrsrc + 3) = b;
				*(ptrsrc + 4) = g;
				*(ptrsrc + 5) = r;
				DMAOR = 0x0001;									/* DMACマスタ:イネーブル */
				SAR_0 = (unsigned long)ptrsrc;					/* 転送元アドレス */
				DAR_0 = (unsigned long)ptrsrc + 6;				/* 転送先アドレス */
				DMATCR_0 = (unsigned long)pheader->size / 2 - 3;/* 転送回数=ワード数(size/2) */
				CHCR_0 = 0x00005429;							/* Dest増加,Src増加,オートリクエスト,バーストモード,16bit幅,INT禁止 */
				while((CHCR_0 & 0x00000002) == 0);		/* 転送終了待ち */
				CHCR_0 &= 0xfffffffc;							/* TE,DEビットクリア */
				DMAOR &= 0xfffe;								/* DMACマスタ:ディセーブル */
#else
				if ((r == 0) && (g == 0) && (b == 0)) {
					memset(ptrsrc, 0x00, pheader->size);	/* DEMO用の高速化:p */
				}
				else {
					for (count = 0; count < pheader->width * pheader->height; count++) {
						*ptrsrc++ = b;
						*ptrsrc++ = g;
						*ptrsrc++ = r;
					}
				}
#endif
				break;
			case LCDC_MODE_RGB565:
				b >>= 3; g >>= 2; r >>= 3;
#ifdef IMG_USE_DMAC
				ptrsrc = (unsigned char *)UNCACHED(ptrsrc);
				if ((pheader->size > 16) && ((pheader->size % 16) == 0)) {
					for (count = 0; count < 8; count++) {
						*((unsigned short *)ptrsrc + count) = (unsigned short)((r << 11) | (g << 5) | b);
					}
					DMAOR = 0x0001;										/* DMACマスタ:イネーブル */
					SAR_0 = (unsigned long)ptrsrc;						/* 転送元アドレス */
					DAR_0 = (unsigned long)ptrsrc + 16;					/* 転送先アドレス */
					DMATCR_0 = (unsigned long)pheader->size / 16 - 1;	/* 転送回数=(size/16) */
					CHCR_0 = 0x00005439;								/* Dest増加,Src増加,オートリクエスト,バーストモード,16バイト転送,INT禁止 */
				}
				else {
					*((unsigned short *)ptrsrc) = (unsigned short)((r << 11) | (g << 5) | b);
					DMAOR = 0x0001;									/* DMACマスタ:イネーブル */
					SAR_0 = (unsigned long)ptrsrc;					/* 転送元アドレス */
					DAR_0 = (unsigned long)ptrsrc + 2;				/* 転送先アドレス */
					DMATCR_0 = (unsigned long)pheader->size / 2 - 1;/* 転送回数=(size/2) */
					CHCR_0 = 0x00005429;							/* Dest増加,Src増加,オートリクエスト,バーストモード,16bit幅,INT禁止 */
				}
				while((CHCR_0 & 0x00000002) == 0);	/* 転送終了待ち */
				CHCR_0 &= 0xfffffffc;						/* TE,DEビットクリア */
				DMAOR &= 0xfffe;							/* DMACマスタ:ディセーブル */
#else
				if ((r == 0) && (g == 0) && (b == 0)) {
					memset(ptrsrc, 0x00, pheader->size);	/* DEMO用の高速化:p */
				}
				else {
					for (count = 0; count < pheader->width * pheader->height; count++) {
						*(unsigned short *)ptrsrc++ = (unsigned short)((r << 11) | (g << 5) | b);
					}
				}
#endif
				break;
			case LCDC_MODE_LUT8:
			default:
				return;
			}
		}
	}
}

void IMG_BoxFill(HIMAGEBUF hBuffer, short x1, short y1, short x2, short y2, unsigned char r, unsigned char g, unsigned char b)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	unsigned char *ptr;
	unsigned short wval, offset;
	short count, temp;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {
			/* 座標のチェック */
			if (((x1 < 0) && (x2 < 0)) || ((x1 >= pheader->width) && (x2 >= pheader->width))) {
				return;
			}
			if (((y1 < 0) && (y2 < 0)) || ((y1 >= pheader->width) && (y2 >= pheader->width))) {
				return;
			}
			if (x1 > x2) {
				temp = x1;
				x1 = x2;
				x2 = temp;
			}
			if (y1 > y2) {
				temp = y1;
				y1 = y2;
				y2 = temp;
			}
			img_checkpos(&x1, &y1, pheader->width, pheader->height);
			img_checkpos(&x2, &y2, pheader->width, pheader->height);

			ptr = (unsigned char *)pheader + sizeof(IMAGEBUF_ST)
				+ (pheader->width * y1 + x1) * pheader->bytesperpixel;
			ptr = (unsigned char *)UNCACHED(ptr);

			offset = pheader->width * pheader->bytesperpixel;

			switch (pheader->mode) {
			case LCDC_MODE_RGB888:
				for (count = 0; count <= y2 - y1; count++) {
					img_hline_rgb888(ptr, x2 - x1 + 1, r, g, b);
					ptr += offset;
				}
				break;
			case LCDC_MODE_RGB565:
				b >>= 3; g >>= 2; r >>= 3;
				wval = ((unsigned short)r << 11) | ((unsigned short)g << 5) | (unsigned short)b;
				for (count = 0; count <= y2 - y1; count++) {
					img_hline_rgb565((unsigned short *)ptr, x2 - x1 + 1, wval);
					ptr += offset;
				}
				break;
			case LCDC_MODE_LUT8:
			default:
				return;
			}
		}
	}
}

void IMG_BitBlt(HIMAGEBUF hBuffer, short x1, short y1, short width, short height, HIMAGEBUF hBufferSrc, short srcx, short srcy)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	PIMAGEBUF_ST phdrsrc = (PIMAGEBUF_ST)hBufferSrc;
	unsigned char *ptrsrc, *ptrdst;
	unsigned short y, count, offset_src, offset_dst;

	if ((pheader != 0) && (phdrsrc != 0)) {
		if ((pheader->magicid == IMAGEBUF_MAGICID) && (phdrsrc->magicid == IMAGEBUF_MAGICID)) {
			/* 座標のチェック */
			if (x1 < 0) {
				width -= (-x1);
				if (width < 0) {
					return;
				}
				srcx  += (-x1);
				x1 = 0;
			}
			else if ((x1 + width) > pheader->width) {
				width = pheader->width - x1;
				if (width <= 0) {
					return;
				}
			}
			else if (x1 >= pheader->width) {
				return;
			}
			else {
				;
			}
			if (y1 < 0) {
				height -= (-y1);
				if (height < 0) {
					return;
				}
				srcy += (-y1);
				y1 = 0;
			} else if ((y1 + height) > pheader->height) {
				height = pheader->height - y1;
				if (height <= 0) {
					return;
				}
			} else if (y1 >= pheader->height) {
				return;
			}
			else {
				;
			}

			ptrdst = (unsigned char *)pheader + sizeof(IMAGEBUF_ST)
					+ (pheader->width * y1 + x1) * pheader->bytesperpixel;
			ptrdst = (unsigned char *)UNCACHED(ptrdst);

			ptrsrc = (unsigned char *)phdrsrc + sizeof(IMAGEBUF_ST)
					+ (phdrsrc->width * srcy + srcx) * phdrsrc->bytesperpixel;
			ptrsrc = (unsigned char *)UNCACHED(ptrsrc);

			offset_src = phdrsrc->width * phdrsrc->bytesperpixel;
			offset_dst = pheader->width * pheader->bytesperpixel;
			count = width * pheader->bytesperpixel;
			for (y = 0; y < height; y++) {
#if 0	/* 未検証 */
				DMAOR = 0x0001;								/* DMACマスタ:イネーブル */
				if (count % 16 == 0) {
					SAR_0 = (unsigned long)ptrsrc;			/* 転送元アドレス */
					DAR_0 = (unsigned long)ptrdst;			/* 転送先アドレス */
					DMATCR_0 = (unsigned long)count / 16;	/* 転送回数=(size/16) */
					CHCR_0 = 0x00005439;					/* Dest増加,Src増加,オートリクエスト,バーストモード,16バイト転送,INT禁止 */
				}
				else {
					SAR_0 = (unsigned long)ptrsrc;			/* 転送元アドレス */
					DAR_0 = (unsigned long)ptrdst;			/* 転送先アドレス */
					DMATCR_0 = (unsigned long)count / 2;	/* 転送回数=(size/2) */
					CHCR_0 = 0x00005429;					/* Dest増加,Src増加,オートリクエスト,バーストモード,16bit幅,INT禁止 */
					while((CHCR_0 & 0x00000002) == 0);		/* 転送終了待ち */
				}
				CHCR_0 &= 0xfffffffc;						/* TE,DEビットクリア */
				DMAOR &= 0xfffe;							/* DMACマスタ:ディセーブル */
#else
				memcpy(ptrdst, ptrsrc, count);
#endif
				ptrsrc += offset_src;
				ptrdst += offset_dst;
			}
		}
	}
}

static void img_checkpos(short *x, short *y, short width, short height)
{
	if (*x < 0) {
		*x = 0;
	}
	else if (*x >= width) {
		*x = width - 1;
	}
	else {
		;
	}

	if (*y < 0) {
		*y = 0;
	}
	else if (*y >= height) {
		*y = height - 1;
	}
	else {
		;
	}
}

void IMG_HLine(HIMAGEBUF hBuffer, short x1, short y1, short x2, unsigned char r, unsigned char g, unsigned char b)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	unsigned char *ptr;
	unsigned short wval;
	short temp;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {

			/* 座標のチェック */
			if (((x1 < 0) && (x2 < 0)) || ((x1 >= pheader->width) && (x2 >= pheader->width))) {
				return;
			}

			if ((y1 < 0) || (y1 >= pheader->height)) {
				return;
			}

			if (x1 > x2) {
				temp = x1;
				x1 = x2;
				x2 = temp;
			}

			img_checkpos(&x1, &y1, pheader->width, pheader->height);
			img_checkpos(&x2, &y1, pheader->width, pheader->height);

			ptr = (unsigned char *)pheader + sizeof(IMAGEBUF_ST)
				+ (pheader->width * y1 + x1) * pheader->bytesperpixel;
			ptr = (unsigned char *)UNCACHED(ptr);

			switch (pheader->mode) {
			case LCDC_MODE_RGB888:
				img_hline_rgb888(ptr, x2 - x1 + 1, r, g, b);
				break;
			case LCDC_MODE_RGB565:
				b >>= 3; g >>= 2; r >>= 3;
				wval = (((unsigned short)r << 11) | ((unsigned short)g << 5) | (unsigned short)b);
				img_hline_rgb565((unsigned short *)ptr, x2 - x1 + 1, wval);
				break;
			case LCDC_MODE_LUT8:
			default:
				return;
			}
		}
	}
}

/* RGB:888フォーマットの横線描画（内部関数） */
static void img_hline_rgb888(unsigned char *ptr, short count, unsigned char r, unsigned char g, unsigned char b)
{
	while (count-- > 0) {
		*ptr++ = (unsigned char)b;
		*ptr++ = (unsigned char)g;
		*ptr++ = (unsigned char)r;
	}
}

/* RGB:565フォーマットの横線描画（内部関数） */
static void img_hline_rgb565(unsigned short *wptr, short count, unsigned short wval)
{
	while (count-- > 0) {
		*wptr++ = wval;
	}
}

void IMG_VLine(HIMAGEBUF hBuffer, short x1, short y1, short y2, unsigned char r, unsigned char g, unsigned char b)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	unsigned char *ptr;
	unsigned short *wptr, wval, offset;
	short count, temp;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {

			/* 座標のチェック */
			if (((y1 < 0) && (y2 < 0)) || ((y1 >= pheader->height) && (y2 >= pheader->height))) {
				return;
			}

			if ((x1 < 0) || (x1 >= pheader->width)) {
				return;
			}

			if (y1 > y2) {
				temp = y1;
				y1 = y2;
				y2 = temp;
			}

			img_checkpos(&x1, &y1, pheader->width, pheader->height);
			img_checkpos(&x1, &y2, pheader->width, pheader->height);

			ptr = (unsigned char *)pheader + sizeof(IMAGEBUF_ST)
				+ (pheader->width * y1 + x1) * pheader->bytesperpixel;
			ptr = (unsigned char *)UNCACHED(ptr);

			switch (pheader->mode) {
			case LCDC_MODE_RGB888:
				offset = pheader->width * pheader->bytesperpixel;
				for (count = 0; count <= y2 - y1; count++) {
					*ptr       = (unsigned char)b;
					*(ptr + 1) = (unsigned char)g;
					*(ptr + 2) = (unsigned char)r;
					ptr += offset;
				}
				break;
			case LCDC_MODE_RGB565:
				wptr = (unsigned short *)ptr;
				b >>= 3; g >>= 2; r >>= 3;
				wval = (((unsigned short)r << 11) | ((unsigned short)g << 5) | (unsigned short)b);
				offset = (pheader->width * pheader->bytesperpixel) / 2;
				for (count = 0; count <= y2 - y1; count++) {
					*wptr = wval;
					wptr += offset;
				}
				break;
			case LCDC_MODE_LUT8:
			default:
				return;
			}
		}
	}
}

void IMG_Rectangle(HIMAGEBUF hBuffer, short x1, short y1, short x2, short y2, unsigned char r, unsigned char g, unsigned char b)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {
			IMG_HLine(hBuffer, x1, y1, x2, r, g, b);
			IMG_HLine(hBuffer, x1, y2, x2, r, g, b);
			IMG_VLine(hBuffer, x1, y1, y2, r, g, b);
			IMG_VLine(hBuffer, x2, y1, y2, r, g, b);
		}
	}
}

#define DIFF(a, b) ((a) < (b) ? (b) - (a): (a) - (b))

void IMG_Line(HIMAGEBUF hBuffer, short x1, short y1, short x2, short y2, unsigned char r, unsigned char g, unsigned char b)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	short dx, dy, sx, sy, err, e2;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {

			dx = DIFF(x1, x2);
			dy = DIFF(y1, y2);

			if (x1 < x2) {
				sx = 1;
			}
			else {
				sx = -1;
			}

			if (y1 < y2) {
				sy = 1;
			}
			else {
				sy = -1;
			}

			err = dx - dy;

			while(1) {
				IMG_SetPixel(hBuffer, x1, y1, r, g, b);
				if ((x1 == x2) && (y1 == y2)) {
					break;
				}
				e2 = 2 * err;
				if (e2 > -dy) {
					err = err - dy;
					x1 += sx;
				}
				if (e2 < dx) {
					err += dx;
					y1 += sy;
				}
			}
		}
	}
}

void IMG_Circle(HIMAGEBUF hBuffer, short cx, short cy, short rn, unsigned char r, unsigned char g, unsigned char b)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	short error = -rn;
	short x = rn;
	short y = 0;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {

			while (x >= y) {

				img_plot4points(hBuffer, cx, cy, x, y, r, g, b);
				if (x != y) {
					img_plot4points(hBuffer, cx, cy, y, x, r, g, b);
				}

				error += y;
				++y;
				error += y;

				if (error >= 0) {
					error -= x;
					--x;
					error -= x;
				}
			}
		}
	}
}

static void img_plot4points(HIMAGEBUF hBuffer, short cx, short cy, short x, short y, unsigned char r, unsigned char g, unsigned char b)
{
	IMG_SetPixel(hBuffer, cx + x, cy + y, r, g, b);
	if (x != 0) {
		IMG_SetPixel(hBuffer, cx - x, cy + y, r, g, b);
	}
	if (y != 0) {
		IMG_SetPixel(hBuffer, cx + x, cy - y, r, g, b);
	}
	if ((x != 0) && (y != 0)) {
		IMG_SetPixel(hBuffer, cx - x, cy - y, r, g, b);
	}
}

void IMG_CircleFill(HIMAGEBUF hBuffer, short cx, short cy, short rn, unsigned char r, unsigned char g, unsigned char b)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	short error = -rn;
	short x = rn;
	short y = 0;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {

			while (x >= y) {

				IMG_HLine(hBuffer, cx - x, cy - y, cx + x, r, g, b);
				IMG_HLine(hBuffer, cx - x, cy + y, cx + x, r, g, b);
				IMG_HLine(hBuffer, cx - y, cy - x, cx + y, r, g, b);
				IMG_HLine(hBuffer, cx - y, cy + x, cx + y, r, g, b);

				error += y;
				++y;
				error += y;

				if (error >= 0) {
					error -= x;
					--x;
					error -= x;
				}
			}
		}
	}
}

void IMG_SetPixel(HIMAGEBUF hBuffer, short x1, short y1, unsigned char r, unsigned char g, unsigned char b)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	unsigned char *ptr;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {

			/* 座標のチェック */
			if ((x1 < 0) || (x1 >= pheader->width)) {
				return;
			}

			if ((y1 < 0) || (y1 >= pheader->height)) {
				return;
			}

			ptr = (unsigned char *)pheader + sizeof(IMAGEBUF_ST)
				+ (pheader->width * y1 + x1) * pheader->bytesperpixel;
			ptr = (unsigned char *)UNCACHED(ptr);

			switch (pheader->mode) {
			case LCDC_MODE_RGB888:
				*ptr++ = (unsigned char)b;
				*ptr++ = (unsigned char)g;
				*ptr++ = (unsigned char)r;
				break;
			case LCDC_MODE_RGB565:
				b >>= 3; g >>= 2; r >>= 3;
				*(unsigned short *)ptr = (((unsigned short)r << 11) | ((unsigned short)g << 5) | (unsigned short)b);
				break;
			case LCDC_MODE_LUT8:
			default:
				return;
			}
		}
	}
}

void IMG_GetPixel(HIMAGEBUF hBuffer, short x1, short y1, unsigned char *pr, unsigned char *pg, unsigned char *pb)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	unsigned char *ptr;
	unsigned short wval;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {

			/* 座標のチェック */
			if ((x1 < 0) || (x1 >= pheader->width)) {
				*pr = *pg = *pb = 0;
				return;
			}

			if ((y1 < 0) || (y1 >= pheader->height)) {
				*pr = *pg = *pb = 0;
				return;
			}

			ptr = (unsigned char *)pheader + sizeof(IMAGEBUF_ST)
				+ (pheader->width * y1 + x1) * pheader->bytesperpixel;
			ptr = (unsigned char *)UNCACHED(ptr);

			switch (pheader->mode) {
			case LCDC_MODE_RGB888:
				*pb = *ptr++;
				*pg = *ptr++;
				*pr = *ptr++;
				break;
			case LCDC_MODE_RGB565:
				wval = *(unsigned short *)ptr;
				*pr = (wval & 0xf800) >> 8;
				*pg = (wval & 0x07e0) >> 3;
				*pb = (wval & 0x001f) << 3;
				break;
			case LCDC_MODE_LUT8:
			default:
				return;
			}
		}
	}
}

void IMG_DrawFont(HIMAGEBUF hBuffer, short x1, short y1, short xmag, short ymag, unsigned char code, unsigned char r, unsigned char g, unsigned char b)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	unsigned short wval, draw_width, draw_height;
	short x, y;
	unsigned char *ptr, *pfnt;
	unsigned char bittbl[IMG_FONT_WIDTH * IMG_FONT_MAXMAG];
	unsigned char bittbl_src[8] = {
		0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
	};

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {

			if ((x1 < 0) || (x1 >= pheader->width)) {
				return;
			}
			if ((y1 < 0) || (y1 >= pheader->height)) {
				return;
			}

			if ((xmag < 1) || (xmag > IMG_FONT_MAXMAG)) {
				return;
			}

			if ((ymag < 1) || (ymag > IMG_FONT_MAXMAG)) {
				return;
			}

			if ((code < 0x20) || (code > 0x7e)) {
				return;
			}

			draw_width = pheader->width - x1;
			if (draw_width > IMG_FONT_WIDTH * xmag) {
				draw_width = IMG_FONT_WIDTH * xmag;
			}

			draw_height = pheader->height - y1;
			if (draw_height > IMG_FONT_HEIGHT * ymag) {
				draw_height = IMG_FONT_HEIGHT * ymag;
			}

			for (x = 0; x < draw_width; x++) {
				bittbl[x] = bittbl_src[x / xmag];
			}

			switch (pheader->mode) {
			case LCDC_MODE_RGB888:
				for (y = 0; y < draw_height; y++) {
					pfnt = (unsigned char *)&g_FontData5x8[code - 0x20][y / ymag];
					ptr = (unsigned char *)pheader + sizeof(IMAGEBUF_ST)
						+ (pheader->width * (y1 + y) + x1) * pheader->bytesperpixel;
					ptr = (unsigned char *)UNCACHED(ptr);
					for (x = 0; x < draw_width; x++) {
						if ((*pfnt & bittbl[x]) != 0) {
							*ptr++ = b;
							*ptr++ = g;
							*ptr++ = r;
						}
						else {
							ptr += 3;
						}
					}
				}
				break;
			case LCDC_MODE_RGB565:
				b >>= 3; g >>= 2; r >>= 3;
				wval = (unsigned short)((r << 11) | (g << 5) | b);
				for (y = 0; y < draw_height; y++) {
					pfnt = (unsigned char *)&g_FontData5x8[code - 0x20][y / ymag];
					ptr = (unsigned char *)pheader + sizeof(IMAGEBUF_ST)
						+ (pheader->width * (y1 + y) + x1) * pheader->bytesperpixel;
					ptr = (unsigned char *)UNCACHED(ptr);
					for (x = 0; x < draw_width; x++) {
						if ((*pfnt & bittbl[x]) != 0) {
							*(unsigned short *)ptr = wval;
						}
						ptr += 2;
					}
				}
				break;
			case LCDC_MODE_LUT8:
				break;
			}
		}
	}
}

void IMG_DrawString(HIMAGEBUF hBuffer, short x1, short y1, short xmag, short ymag, char *pstring, unsigned char r, unsigned char g, unsigned char b)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	int count;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {
			for (count = 0; count < strlen(pstring); count++) {
				IMG_DrawFont(hBuffer, x1, y1, xmag, ymag, *(pstring + count), r, g, b);
				x1 += (IMG_FONT_WIDTH + 1) * xmag;
			}
		}
	}
}

void IMG_Scroll(HIMAGEBUF hBuffer, short h, short v)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	short x, y, width, height, srcx, srcy;

	if ((h == 0) && (v == 0)) {
		return;
	}

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {
			if (v > 0) {
				y = 0;
				height = pheader->height - v;
				srcy = v;
			}
			else {
				y = -v;
				height = pheader->height + v;
				srcy = 0;
			}
			
			if (h > 0) {
				x = 0;
				width = pheader->width - h;
				srcx = h;
			}
			else {
				x = -h;
				width = pheader->width + h;
				srcx = 0;
			}
			IMG_BitBlt(hBuffer, x, y, width, height, hBuffer, srcx, srcy);
		}
	}
}

unsigned char* IMG_GetBits(HIMAGEBUF hBuffer, unsigned short option)
{
	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;
	unsigned char *pbuf = 0;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {
			if (option & IMG_GETBITS_READONLY) {
				pbuf = (unsigned char *)pheader + sizeof(IMAGEBUF_ST);
			}
			else {
				pbuf = (unsigned char *)UNCACHED((unsigned char *)pheader + sizeof(IMAGEBUF_ST));
			}
		}
	}

	return pbuf;
}

unsigned short IMG_GetWidth(HIMAGEBUF hBuffer)
{
	unsigned short result = 0;

	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {
			result = pheader->width;
		}
	}

	return result;
}
unsigned short IMG_GetHeight(HIMAGEBUF hBuffer)
{
	unsigned short result = 0;

	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {
			result = pheader->height;
		}
	}

	return result;
}

unsigned short IMG_GetBitsPerPixel(HIMAGEBUF hBuffer)
{
	unsigned short result = 0;

	PIMAGEBUF_ST pheader = (PIMAGEBUF_ST)hBuffer;

	if (pheader != 0) {
		if (pheader->magicid == IMAGEBUF_MAGICID) {
			result = pheader->bitsperpixel;
		}
	}

	return result;
}
